#define SystemSevenOrLater 1

#include <Carbon/Carbon.r>

resource 'STR#' (127)
{{
    "DjVu File",
    "DjVu File",
    "DjVu File",
    "DjVu File (obsolete)",
    "DjVu File (obsolete)",
    "DjVu File"
}};

resource 'STR#' (128)
{{
    "image/x-djvu", "djvu,djv",
    "image/x.djvu", "",
    "image/djvu", "",
    "image/x-dejavu", "",
    "image/x-iw44", "",
    "image/vnd.djvu", ""
}};

resource 'STR#' (126)
{{
  "The DjView4 browser plugin"
}};
