
1 ICON DISCARDABLE "../src/images/DjVuApp.ico"

1 VERSIONINFO
 FILEVERSION 1,0,0,1
 PRODUCTVERSION 1,0,0,1
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x4L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904e4"
        BEGIN
            VALUE "Comments", "\0"
            VALUE "CompanyName", "\0"
            VALUE "FileDescription", "npdjvu\0"
            VALUE "FileOpenName", "DjVu File|DjVu File|DjVu File|DjVu File (obsolete)|DjVu File (obsolete)|DjVu File"
            VALUE "FileExtents", "djvu,djv|||||\0"
            VALUE "FileVersion", "1, 0, 0, 1\0"
            VALUE "InternalName", "npdjvu\0"
            VALUE "LegalCopyright", "\0"
            VALUE "LegalTrademarks", "\0"
            VALUE "MIMEType", "image/x-djvu|image/x.djvu|image/djvu|image/x-dejavu|image/x-iw44|image/vnd.djvu\0"
            VALUE "PrivateBuild", "\0"
            VALUE "ProductName", "DjView4 browser plugin\0"
            VALUE "ProductVersion", "1, 0, 0, 1\0"
            VALUE "SpecialBuild", "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1252
    END
END

