#ifndef VERSION_H
#define VERSION_H

#ifndef DJVIEW_ORG
# define DJVIEW_ORG "DjVuLibre"
#endif
#ifndef DJVIEW_DOMAIN
# define DJVIEW_DOMAIN "djvu.org"
#endif
#ifndef DJVIEW_APP
# define DJVIEW_APP "DjView"
#endif
#ifndef DJVIEW_VERSION
# define DJVIEW_VERSION 0x40c00
# define DJVIEW_VERSION_STR "4.12"
# define RC_VERSION 4,12,0,0
# define RC_VERSION_STR "4,12,0,0\0"
#endif

#endif

